#ifndef OBJECTS_H
#define OBJECTS_H

//-----------------------------------------------------------------------------
#include "Base.h"
#include <limits>

#ifdef max
#undef max
#undef min
#endif

using namespace std;

//-----------------------------------------------------------------------------
class CScene;
class CObject;
class CLight;

//-----------------------------------------------------------------------------
/// ray class
/*
* 
*/
class CRay
{
//-----------------------------------------------------------------------------

  /** source object */
  const CObject* sO;
  /** start point */
  const Point3d p;
  /** direction */
  const Point3d d;

//-----------------------------------------------------------------------------
public:

  /** Ctor. */
  CRay( const Point3d& start, const Point3d& direction, 
        const CObject* source = NULL )
  : sO( source ), p( start ), d( direction )
  {};

  /** Get the start point of the ray. */
  const Point3d& start() const {
    return p;
  };

  /** Get the ray direction. */
  const Point3d& dir() const {
    return d;
  };

  /** Get the source object of the ray. */
  const CObject* source() const {
    return sO;
  };
};


//-----------------------------------------------------------------------------
/// basic object class
/*
* 
*/
class CObject
{
//-----------------------------------------------------------------------------
protected:

  /** position */
  Point3d p;

  /** color */
  CColor c;

  /** reflection coefficient */
  double reflect;

  /** refraction coefficient */
  double refract;

//-----------------------------------------------------------------------------
public:
 
  /** 
  * Ctor. 
  * @param pos - object position
  * @param color - object color
  * @param reflect - reflection coefficient
  * @param refract - refraction coefficient
  */
  CObject( const Point3d& pos, const CColor& color, 
    const double reflect, const double refract )
    : p( pos ), c( color ), reflect( reflect ), refract( refract )
  {};

  /** Get the object position (const). */
  const Point3d& pos() const 
  {
    return p;
  };

  /** Get the object position. */
  Point3d& pos()
  {
    return p;
  };

  /** Get the object color. */
  const CColor& color() const
  {
    return c;
  };

  /** Get the reflection coefficient. */
  const double reflection() const
  {
    return reflect;
  };

  /** Get the refraction coefficient. */
  const double refraction() const
  {
    return refract;
  };

  // change reflection/refraction coefficient by delta; clamp the result into the range [0,1]
  void changeReflection( double delta );
  void changeRefraction( double delta );

  /** Get the intersection point between this object and the given ray. */
  virtual const Point3d intersection( const CRay& ) const 
  {
    return Point3d( -numeric_limits<double>::max() );
  };

  /** Get the normal at the given point. */
  virtual const Point3d normalAt( const Point3d& ) const
  {
    return Point3d();
  };

  /** Get the color at the given point. */
  virtual const CColor colorAt( const Point3d&, 
							    const Point3d&, const Point3d&, const CLight& ) const
  {
    return c;
  };

};


//-----------------------------------------------------------------------------
/// light class
/*
* 
*/
class CLight : public CObject
{
//-----------------------------------------------------------------------------
public:
 
  /** 
  * Ctor. 
  * @param pos - light position
  * @param color - light color
  */
  CLight( const Point3d& pos, const CColor& color )
    : CObject( pos, color, 0, 0 )
  {};

};


//-----------------------------------------------------------------------------
/// sphere class
/*
* 
*/
class CSphere : public CObject
{
  /** radius^2 */
  const double r2;

//-----------------------------------------------------------------------------
public:
 
  /** 
  * Ctor. 
  * @param pos - sphere position
  * @param radius - the sphere radius
  * @param color - sphere color
  * @param reflect - reflection coefficient
  * @param refract - refraction coefficient
  */
  CSphere( const Point3d& pos, const double radius, const CColor& color, 
    const double reflect, const double refract )
  : CObject( pos, color, reflect, refract ), r2( radius*radius )
  {};

  /** Get the intersection point between this object and the given ray. */
  const Point3d intersection( const CRay& ray ) const;

  /** Get the normal at the given point. */
  const Point3d normalAt( const Point3d& iP ) const;

  /**
  * Get the color at the given point. 
  * @param iP - intersection point
  * @param n - the normal at the intersection point
  * @param l - the light direction vector
  * @param light - the light
  */
  const CColor colorAt( const Point3d& iP, const Point3d& n,
    const Point3d& l, const CLight& light ) const;
};


//-----------------------------------------------------------------------------
/// plane class
/*
* 
*/
class CPlane : public CObject
{
  /** the plane normal */
  const Point3d normal;
  /** distance to the origin */
  const double d;

//-----------------------------------------------------------------------------
public:
 
  /** 
  * Ctor. 
  * @param pos - sphere position
  * @param radius - the sphere radius
  * @param color - sphere color
  * @param reflect - reflection coefficient
  * @param refract - refraction coefficient
  */
  CPlane( const Point3d& p0, const Point3d& normal, 
    const CColor& color, const double refCoeff, const double refractCoeff )
  : CObject( p0, color, refCoeff, refractCoeff ), normal( normal ),
    d( p0 * normal )
  {};

  /** Get the intersection point between this object and the given ray. */
  const Point3d intersection( const CRay& ray ) const;

  /** Get the normal at the given point. */
  const Point3d normalAt( const Point3d& iP ) const;

  /**
  * Get the color at the given point. 
  * @param iP - intersection point
  * @param n - the normal at the intersection point
  * @param l - the light direction vector
  * @param light - the light
  */
  const CColor colorAt( const Point3d& iP, const Point3d& n,
    const Point3d& l, const CLight& light ) const;
};

#endif
