//-----------------------------------------------------------------------------
// local includes
#include "MainWindow.h"
#include "Objects.h"
#include "Scene.h"

#include <math.h>

//-----------------------------------------------------------------------------
using namespace std;


//-----------------------------------------------------------------------------
MainWindow::MainWindow( QWidget* parent, Qt::WindowFlags flags )
: QMainWindow( parent, flags ), scene( NULL ), tau( 0 ), zPos( 0.0 ), pause( false )
{
  // initialize the scene
  initScene();
}


//-----------------------------------------------------------------------------
MainWindow::~MainWindow()
{
  // delete the scene
  if ( scene )
    delete scene;
}


//-----------------------------------------------------------------------------
//void MainWindow::keyReleaseEvent( QKeyEvent* e )
void MainWindow::keyPressEvent( QKeyEvent* e )
{
  // pause
  if ( e->key() == Qt::Key_P || e->key() == Qt::Key_Space )
    pause = !pause;

  // close
  if ( e->key() == Qt::Key_Escape )
    close();

  // save the image
  if ( e->key() == Qt::Key_S )
    save();

  // increase raytracing depth
  if ( e->key() == Qt::Key_Plus || e->key() == Qt::Key_Equal )
    scene->increaseRenderDepth();
  // decrease raytracing depth
  if ( e->key() == Qt::Key_Minus )
    scene->decreaseRenderDepth();

  // go into screen
  if ( e->key() == Qt::Key_Up )
  {
    zPos += 1;
    scene->setNearClippingPlane( zPos );
  }
  // go out off screen
  if ( e->key() == Qt::Key_Down )
  {
    zPos -= 1;
    scene->setNearClippingPlane( zPos );
  }

  // next frame
  if ( e->key() == Qt::Key_Right )
    tau += 0.1;
  // previous frame
  if ( e->key() == Qt::Key_Left )
    tau -= 0.1;

	// change refraction / reflection of glass sphere
	if ( e->key() == Qt::Key_Less || e->key() == Qt::Key_Comma )
		scene->getObject(3)->changeRefraction( -0.1 );
	if ( e->key() == Qt::Key_Greater || e->key() == Qt::Key_Period )
		scene->getObject(3)->changeRefraction(  0.1 );
	if ( e->key() == Qt::Key_Z )
		scene->getObject(3)->changeReflection( -0.1 );
	if ( e->key() == Qt::Key_X )
		scene->getObject(3)->changeReflection(  0.1 );
	
	if ( pause )
		render();
}


//-----------------------------------------------------------------------------
void MainWindow::resizeEvent( QResizeEvent* )
{
  // create a new image with the new size
  image = QImage( width(), height(), QImage::Format_RGB32 );

  // render the scene
  render();
}


//-----------------------------------------------------------------------------
void MainWindow::save()
{
  // get the file name
  const QString& fname = QFileDialog::getSaveFileName( this, "Save File",
    ".", "Image Files ( *.png *.jpg *.jpeg *.tiff )" );

  if ( fname.isNull() || fname.isEmpty() )
    return;

  // save the image
  image.save( fname );
}


//-----------------------------------------------------------------------------
void MainWindow::timerEvent( QTimerEvent* )
{
  // process qt events
  qApp->processEvents();

  // check for pause
  if ( pause )
    return;

  // time update
  tau += 0.1;

  // draw the scene
  render();
}


//-----------------------------------------------------------------------------
void MainWindow::paintEvent( QPaintEvent* )
{
  // draw the saved image
  QPainter painter( this );
  painter.drawImage( 0, 0, image );
}


//-----------------------------------------------------------------------------
void MainWindow::initScene()
{
  // create a new scene
  scene = new CScene( CColor( 0, 100, 255 ), zPos, 20.0 );

  // create a sphere
  scene->addObject( new CSphere( Point3d( -1.5, 0, 5 ), 1.0, CColor( 255, 0, 0 ),   0.7, 0 ) );
  scene->addObject( new CSphere( Point3d( 1.5, 1, 6 ),  1.0, CColor( 0, 255, 0 ),   0.7, 0 ) );
  scene->addObject( new CSphere( Point3d( 0, 0, 8 ),    1.0, CColor( 255, 0, 255 ), 0.7, 0 ) );

  // yellow glass sphere
  scene->addObject( new CSphere( Point3d( 0.9, 0, 4 ), 0.5, CColor( 255, 255, 100 ), 0.2, 0.8 ) );

  // add a ground plane
  scene->addObject( new CPlane( Point3d( 0, -2, 0 ), Point3d( 0, 1, 0 ),
    CColor( 0, 0, 0 ), 0.5, 0.0 ) );

  // 1st light source
  scene->addLight( new CLight( Point3d( 0, 5, 5 ), CColor( 255, 255, 255 ) ) );
}


//-----------------------------------------------------------------------------
void MainWindow::render()
{ 
  // start a timer for fps recording
  Timer t;
  t.start();

  // width and height of the window
  const int w = image.width();
  const int h = image.height();

  // rotate the 1st light
  scene->light( 0 ).pos() = Point3d( 4*sin( tau ), 5, 5.0 + 4*cos( tau ) );
  
  // render the image
  #pragma omp parallel
  {
    #pragma omp for
    for ( int j = 0; j < h; ++j )
    {
      // actual image line
      QRgb* scanLine = (QRgb*)image.scanLine( j );

      // y direction of the rays through the actual line
      const double lineY = ( h-1 - 2.0*j ) / w;

      for ( int i = 0; i < w; ++i )
      {
        // ray direction
        const Point3d dir( -1 + (2.0*i) / w, lineY, 1.0 );

        // ray from viewer
        CRay ray( Point3d( 0, 0, zPos ), dir.normalized() );

        // set the pixel color
        scanLine[ i ] = scene->intersection( ray ).qRgb();
      }
    }
  }

  // set fps timing
  t.stop();
  setWindowTitle( QString( "Raytracing with recursion depth = %1 at %2 FPS" ).arg( 
    scene->maxRenderDepth() ).arg( 1 / t.getElapsedTimeInSec() ) );

  // repaint the image
  repaint();
}
